/* Function imagedoc */

extern imagedoc;
/* DOCUMENT ****************************************************
 *
 *      Package by Niels J. Westergaard 2003-03-30 for
 *      image analysis.
 *      2008-09-24/NJW added palette manipulation functions

    _colscale           getimval            peaksearch1         
    _pal_write          getimvals           peaksearch2         
    add_peak            gkernel             rebin               
    convol2d            image2events        remap               
    cur_extract_box     imcolorbar          signif              
    curmark             imcut               src_cts             
    curmark1            imhcut              unwiden_image       
    disp                mk_palette          w0                  
    disp9               pal                 w1                  
    dispc               pal_edit            w2                  
    distances           pal_plot            w3                  
    edge_suppress       palastro            w4                  
    flatfield           peaksearch0         widen_image         
    manip_image_get     manip_image_put

*****************************************************/

/* Function mk_palette */

func mk_palette( name )
/* DOCUMENT mk_palette, name

   The RGB are defined individually by mouse clicking in coordinate system.
   The argument 'name' is without the '.gp' extension.

   2008-10-10/NJW (or about that time)

*/
{
   if( typeof(name) != "string" ) error,"A name must be given";
   window,0;
   ncolors = 240;
   // define red
   plot,[1,ncolors],[0,0],yr=[0,256];
   oplot,[1,ncolors],[255,255];
   oplot,[1,1],[0,255];
   oplot,[ncolors,ncolors],[0,255];
   write,"Define the RED color function :";
   r = _colscale( ncolors );
   oplot, r, color="red";
   write,"Define the GREEN color function :";
   g = _colscale( ncolors );
   oplot, g, color="green";
   write,"Define the BLUE color function :";
   b = _colscale( ncolors );
   oplot, b, color="blue";

   _pal_write, name, r, g, b;

   write,format="Palette now stored in %s.gp\n", name;

}

/* Function _pal_write */

func _pal_write( name, r, g, b )
{
   ncolors = numberof(r);
   stream = open(GISTDIR+name+".gp","w");
   write,stream,format="# Palette created by mk_palette %s/NJW\n", ndate(3);
   write,stream,format="ncolors= %i\n", ncolors;
   write,stream,format="#    r   g    %s\n","b";
   for( i = 1; i <= ncolors; i++ ) {
      write,stream,format="%8i%8i%8i\n", r(i), g(i), b(i);
   }
   close, stream;

}

/* Function _colscale */

func _colscale( ncolors )
{
   rc = curmark();
   n = numberof(rc)/2;
   rcx = rc(1:0:2);
   rcy = rc(2:0:2);
   is = sort(rcx);
   rcx = rcx(is);
   rcy = rcy(is);
   rcx(1) = 1.;
   if( rcx(2) < rcx(1) ) rcx(2) = 2.;
   rcx(0) = ncolors;
   if( rcx(-1) > rcx(0) ) rcx(-1) = ncolors-1;
   r = long(interp( rcy, rcx, indgen(ncolors) ) );
   w = where( r < 0 );
   if( numberof(w) ) r(w) = 0;
   w = where( r > 255 );
   if( numberof(w) ) r(w) = 255;

   return r;

}

/* Function pal_edit */

func pal_edit( name, step= )
{
   if( is_void(step) ) step = 10;

   p = GISTDIR+name+".gp";
   pal_plot, name;
   r = rscol(p,1,lng=1,silent=1,nomem=1);
   g = rscol(p,2,lng=1,silent=1,nomem=1);
   b = rscol(p,3,lng=1,silent=1,nomem=1);
   ncolors = numberof(r);
   rgb = array(long,ncolors,3);
   rgb(,1) = r;
   rgb(,2) = g;
   rgb(,3) = b;

   ans = "";
   answ = ["r","g","b"];
   cstr = ["red","green","blue"];
   read,prompt="Enter color (r,g, or b) : ... ", ans;
   w = where( ans == answ );
   if( !numberof(w) ) error,"Illegal input";
   nc = w(1);
   again = 1;
   h = 1;
   grow,h,indgen(step:ncolors:step);

   //while( again ) {
      oplot,h,rgb(h,nc),ps=2,symsize=2,color=cstr(nc);
      for( i = 1; i <= numberof(h); i++ ) {
         oplot,[h(i),h(i)],[0,255],color=cstr(nc);
      }
      pts = curmark();
      npts = numberof(pts);
      xpts = pts(1:npts:2);
      ypts = pts(2:npts:2);
      npts /= 2;
      y0 = rgb(h,nc);
      for( i = 1; i <= npts; i++ ) {
         w = where( min(abs(xpts(i)-h)) == abs(xpts(i)-h) )(1);
         y0(w) = long(ypts(i));
      }
      rgb(,nc) = long(interp( y0, h, indgen(ncolors) ));
      w = where( rgb(,nc) < 0 ); if(numberof(w)) rgb(w,nc) = 0;
      w = where( rgb(,nc) > 255 ); if(numberof(w)) rgb(w,nc) = 255;
      oplot, rgb(,nc), color=cstr(nc);
      _pal_write, name, rgb(,1), rgb(,2), rgb(,3);
      pal_plot, name;
      again = 0;
   //}
}


/* Function pal_plot */

func pal_plot( name )
{
   pal, name;
   p = GISTDIR+name+".gp";
   r = rscol(p,1,lng=1,silent=1,nomem=1);
   g = rscol(p,2,lng=1,silent=1,nomem=1);
   b = rscol(p,3,lng=1,silent=1,nomem=1);
   ncolors = numberof(r);
   a = array(double,ncolors,256);
   for(i=1;i<=ncolors;i++) a(i,) = i-1;
   disp,a,reset=1,title=name+".gp";

   oplot,r,color="red";
   oplot,g,color="green";
   oplot,b,color="blue";

}

/* Function flatfield */

func flatfield( image, sigma, sel= )
/* DOCUMENT new_image = flatfield( image, sigma, sel= )

   The image 'image' is folded with a gaussian kernel
   with sigma = 'sigma' and then subtracted from the
   original image. The process is repeated with
   excluded source areas.
   Keyword 'sel' must be an array of same dimension
   as 'image' with 1 where pixels are useful and zero
   elsewhere.

   2007-06-19/NJW

*/
{
   require, "fconvol.i";

   di = dimsof(image);
   if( di(1) != 2 ) {
      write,"Not a 2D image";
      return [];
   }
   dosel = 0;
   if( !is_void(sel) ) {
      d = dimsof(sel);
      if( d(1) != 2 ) {
         write,"'sel' is not a 2D image";
         return [];
      }
      if( d(2) != di(2) || d(3) != di(3) ) {
         write,"'sel' is not commensurate with 'image'";
         return [];
      }
      dosel = 1;
   }
   w = dosel ? where( sel ) : indgen(di(2)*di(3));
   //   Avoid changing the input image
   im1 = image;
   fimage = gfconvol( image, sigma );
   //    First subtraction
   diff  = image - fimage;
   //    Substitute image parts that exceed 3sigma
   //    to avoid subtracting too much around sources
   rms_1 = wrms( diff(w) );
   q = where( abs(diff) > 3*rms_1 );
   if( numberof(q) > 0 ) {
      im1(q) = fimage(q);
      fimage = gfconvol( im1, sigma );
      diff = image - fimage;
   }

   return diff;
}


/* Function src_cts */

func src_cts( image )
/* DOCUMENT number_of_counts = src_cts( image )

   Interactive function to get the number of source
   counts under the assumption that 'image' is a 2D
   array displayed in an active plot window where the
   unit of the axes is in pixels.
   'curmark1' is called twice; first to define the
   source area and then to define the background area.

   The source area must either be completely included
   in the background area or the two areas should be
   disjunct.

   2007-02-06/NJW
*/
{

   p_src = curmark1( style=1, prompt="Mark source ..." );
   p_bkg = curmark1( style=1, prompt="Mark background ..." );
   if( p_src(1) > p_src(3) ) { // swap
      x = p_src(3); p_src(3) = p_src(1); p_src(1) = x;
   }
   if( p_src(2) > p_src(4) ) { // swap
      x = p_src(4); p_src(4) = p_src(2); p_src(2) = x;
   }
   if( p_bkg(1) > p_bkg(3) ) { // swap
      x = p_bkg(3); p_bkg(3) = p_bkg(1); p_bkg(1) = x;
   }
   if( p_bkg(2) > p_bkg(4) ) { // swap
      x = p_bkg(4); p_bkg(4) = p_bkg(2); p_bkg(2) = x;
   }
   dms = dimsof( image );
   x = span(1,dms(2),dms(2))(,-:1:dms(3));
   y = span(1,dms(3),dms(3))(-:1:dms(2),);
   i_src = x > p_src(1) & x < p_src(3) & y > p_src(2) & y < p_src(4);
   w_src = where( i_src );
   nw_src = numberof( w_src );
   i_bkg = x > p_bkg(1) & x < p_bkg(3) & y > p_bkg(2) & y < p_bkg(4);
   w_bkg = where ( i_bkg & (!i_src) );
   nw_bkg = numberof( w_bkg );

   if( nw_src == 0 || nw_bkg == 0 ) {
      write,"Source or background area is zero!";
      return [];
   }
   bkg_level = avg( image(w_bkg) );
   write,format="Number of src pixels: %i\n", nw_src;
   write,format="Number of bkg pixels: %i\n", nw_bkg;
   write,format="Bkg level: %f\n", bkg_level;
   write,format="Cts in source area: %f\n", sum(image(w_src));
   src_counts = sum(image(w_src)) - bkg_level * nw_src;
   write,format="Src counts: %f\n", src_counts;

   return src_counts;

}

/* Function palastro */

func palastro(num)
/* DOCUMENT palastro[, num]

   defines the astro-num palette for plot window

    2007-02-06/NJW
*/
{
   require, "idlx.i";

   if( is_void(num) ) num = 1;
   snum = itoa(num);
   palette,GISTDIR+"astro"+snum+".gp";
}

/* Function pal */

func pal(name)
/* DOCUMENT pal, name

   defines the named palette for plot window

    2007-03-05/NJW, 2010-04-20/NJW update
*/
{
   require, "idlx.i";
   local dirname, basname;

   if( is_void(name) ) {
      write,format="GISTDIR: %s\n", GISTDIR;
      list = file_search("*.gp",GISTDIR);
      for( i = 1; i <= numberof(list); i++ ) {
         splitfname, list(i), dirname, basname;
         write,format="%20s\n",basname;
      }
      return;
   }
   palette,GISTDIR+"/"+name+".gp";
}

func w0(a) { window,0; }
func w1(a) { window,1; }
func w2(a) { window,2; }
func w3(a) { window,3; }
func w4(a) { window,4; }

/* Function 'distances' */

func distances( arg1, arg2, arg3, arg4, arg5, arg6 )
/* DOCUMENT array = distances( dim1, cen1 )
                     1 D version
            array = distances( dim1, dim2, cen1, cen2 )
                     2 D version
            array = distances( dim1, dim2, dim3, cen1, cen2, cen3 )
                     3 D version
      returns an array with dimensions dim1, dim2, ... with
      distances in units of pixel
      from the position (cen1, cen2, ...)

   To get a symmetrical image: cen = double(dim + 1)/2.

*/

{
   if( is_void(arg3) ) { // 1D version
      dim1 = arg1; cen1 = arg2;
      return abs(indgen(dim1) - cen1);
   }
   if( is_void(arg5) ) { // 2D version
      dim1 = arg1; dim2 = arg2;
      cen1 = arg3; cen2 = arg4;
      x = indgen(dim1)(,-:1:dim2);
      y = indgen(dim2)(-:1:dim1,);
      return sqrt((x-cen1)^2 + (y-cen2)^2);
   }
   // 3D version
   dim1 = arg1; dim2 = arg2; dim3 = arg3;
   cen1 = arg4; cen2 = arg5; cen3 = arg6;
   x = indgen(dim1)(,-:1:dim2,-:1:dim3);
   y = indgen(dim2)(-:1:dim1,,-:1:dim3);
   z = indgen(dim3)(-:1:dim1,-:1:dim2,);
   return sqrt((x-cen1)^2 + (y-cen2)^2 + (z-cen3)^2);
}

/* Function disp */

func disp( im, pane=, reset=, xax=, yax=, title=, xtitle=, \
           ytitle=, lcut=, ucut=, Lcut=, Ucut=, pal=, uplf=, cb= )
/* DOCUMENT disp, im, pane=, reset=, xax=, yax=, title=, xtitle=, \
                      ytitle=, lcut=, ucut=, Lcut=, Ucut=, pal=, uplf=, cb=

      Display an image on the screen

              pane : Plot window number (0 thru 7), default is current
     Keywords lcut : Lower cut
              ucut : Upper cut, values in accumulated distribution
                Both defined as fractions compared to the interval
                min(im) - max(im)
              Ucut : Upper cut in image values
              Lcut : Lower cut in image values
              pal  : Name of palette to use (.gp name extension is added if missing)
              cb   : Add a colorbar with this number of ticbars

     2007-02-06/NJW
     2009-11-02/NJW Using 'pli' iso. 'plf'
*/

{
   local x, h;

   if( !is_void(pane) ) window, pane;
   if( !is_void(pal) ) {
      if( strpart(pal,-2:0) != ".gp" ) pal += ".gp";
      palette, pal;
   }

  /*
   *   Apply cuts if defined
   */

   im_work = im;
   if( !is_void(lcut) || !is_void(ucut) ) {
      w = where( im_work != 0 );
      if( numberof(w) > 0 ) {
         binsize = (max(im_work) - min(im_work)) / 1000.0;
         histos, im_work(w), h, x, binsize=binsize;
         h = arr_accum(h, norm=1);

         if( !is_void(lcut) ) {
            index1 = long(interp(indgen(numberof(h)),h,lcut));
            val1 = x(index1);
            w1 = where( im_work < val1 );
            if( numberof(w1) > 0 ) im_work(w1) = val1;
         }

         if( !is_void(ucut) ) {
            index2 = long(interp(indgen(numberof(h)),h,ucut));
            val2 = x(index2);
            w2 = where( im_work > val2 );
            if( numberof(w2) > 0 ) im_work(w2) = val2;
         }
      }
   }
   if( !is_void(Ucut) ) {
      w = where( im_work > Ucut );
      if( numberof(w) ) im_work(w) = Ucut;
      else im_work(0) = Ucut;// statement inserted 2011-02-08/NJW
   }
   if( !is_void(Lcut) ) {
      w = where( im_work < Lcut );
      if( numberof(w) ) im_work(w) = Lcut;
      else im_work(1) = Lcut;// statement inserted 2011-02-08/NJW
   }
   sz = dimsof(im_work);
   if( sz(1) == 2 ) {
      nxax = numberof(xax);
      nyax = numberof(yax);
      if( nxax > 0 ) {
         if( nxax != sz(2) ) {
            write,"Bad dimension of xax keyword";
            return;
         }
         f = double(sz(2))/(sz(2)-1);
         c = 0.5*(xax(0)+xax(1));
         d = 0.5*(xax(0)-xax(1));
         xv = span(c-f*d, c+f*d,sz(2)+1);
      } else {
         xv = span(0.5,sz(2)+0.5,sz(2)+1);
      }
      if( nyax > 0 ) {
         if( nyax != sz(3) ) {
            write,"Bad dimension of yax keyword";
            return;
         }
         f = double(sz(2))/(sz(2)-1);
         c = 0.5*(yax(0)+yax(1));
         d = 0.5*(yax(0)-yax(1));
         yv = span(c-f*d, c+f*d,sz(3)+1);
      } else {
         yv = span(0.5,sz(3)+0.5,sz(3)+1);
      }
      x = xv(,-:1:sz(3)+1);
      y = yv(-:1:sz(2)+1,);
      fma;
      limits; // this statement added 2009-03-13/NJW
      if( uplf ) {
         plf,im_work,y,x,legend="";
      } else {
         pli,im_work,xv(1),yv(1),xv(0),yv(0),legend="";
      }
      if( reset ) limits,"e","e","e","e";
      if( !is_void(title) ) pltitle, title;
      if( is_void(xtitle) ) xtitle="";
      if( is_void(ytitle) ) ytitle="";
      xytitles, xtitle, ytitle;
      plotcomments,init=1; // this statement added 2009-03-13/NJW
      if( cb ) colorbar,sigdig(min(im_work),3),sigdig(max(im_work),3),levs=cb;
   } else {
      write,"Not a 2D array";
   }
}


/* Function dispc */

func dispc( im, pane=, xax=, yax=, levels=, itype=, color=, over=, title=, xtitle=, ytitle= )
/* DOCUMENT dispc, im, pane=, xax=, yax=, levels=, itype=, color=, over=,
                     title=, xtitle=, ytitle=
      Display a contour plot on the plot window
      Keywords:
          pane: plot window number
          itype: axis transforms 0: both linear, 1: y log, x lin,  2: y lin, xlog, 3: both log
          xax: x-axis coordinate values
          yax: y-axis coordinate values
          levels: array of level values
          color: standard meaning
          over: causes overplot
*/

{
   if( !is_void(pane) ) window, pane;

   sz = dimsof(im);
   if( sz(1) != 2 ) {
      write,"Not a 2D array";
      return;
   }
   if( is_void(levels) ) {
      levels=span(min(im),max(im),6)(zcen);
   } else if( numberof(levels) == 1 ) {
         levels=span(min(im),max(im),long(levels)+1)(zcen);
   }
   nxax = numberof(xax);
   nyax = numberof(yax);
   if( nxax > 0 ) {
      if( nxax != sz(2) ) {
         write,"Bad dimension of xax keyword";
         return;
      }
      xv = xax;
   } else {
      xv = indgen(sz(2));
   }
   if( nyax > 0 ) {
      if( nyax != sz(3) ) {
         write,"Bad dimension of yax keyword";
         return;
      }
      yv = yax;
   } else {
      yv = indgen(sz(3));
   }
   x = xv(,-:1:sz(3));
   y = yv(-:1:sz(2),);
   fma;
   if( !is_void(itype) ) mlogxy, itype > 1, itype%2;
   limits, xv(1), xv(0), yv(1), yv(0);
   if( over) {
      plfc,im,y,x,levs=levels;
   }
   go_ahead = 1;
   if( !is_void(color) ) if( color == 0 ) go_ahead = 0;
   if( go_ahead ) plc,im,y,x,levs=levels,marks=0,color=color,legend="";
   if( !is_void(title) ) pltitle, title;
   if( is_void(xtitle) ) xtitle="";
   if( is_void(ytitle) ) ytitle="";
   xytitles, xtitle, ytitle;
   plotcomments,init=1; // this statement added 2009-11-04/NJW
}

func dispcc( im, xax, yax, levels= )
{
   plot,xax,yax,itype=2;
   dms = dimsof(im);
   x = xax(,-:1:dms(3));
   y = yax(-:1:dms(2),);
   //+ limits, 1, dms(2), 1, dms(3);
   plfc,im,y,x,levs=levels;
}

/* Function add_peak */

func add_peak( im, xp, yp, sigmax, sigmay, peak=, volume= )
/* DOCUMENT add_peak( im, xp, yp, sigmax[, sigmay], peak=, volume= )

   Returns the image with a gaussian peak added.
   Coordinate (i,j) refers to center of pixel (i,j).
   If 'sigmay' is not given it is assumed equal to 'sigmax'.
   Keyword   peak   :  Defining peak value
             volume :  Volume of peak (defaults to 1)
                       If given it will override the 'peak' value

   2006-02-23/NJW (Updated from earlier version ca. 2004)
   2010-04-06/NJW Updated to use proper integral over pixels
                  and keywords
*/

{

   if( !is_void(volume) ) {
      peak = [];
   } else {
      volume = 1.;
   }
   if( is_void(sigmay) ) sigmay = sigmax;
   sx = abs(sigmax) * sqrt(2);
   sy = abs(sigmay) * sqrt(2);
   dms = dimsof(im);
   if( dms(1) == 2 ) {
      ii = indgen(dms(2));
      arg1 = (ii + 0.5 - xp)/sx;
      arg2 = (ii - 0.5 - xp)/sx;
      tx = erf(arg1) - erf(arg2);
      jj = indgen(dms(3));
      arg1 = (jj + 0.5 - yp)/sy;
      arg2 = (jj - 0.5 - yp)/sy;
      ty = erf(arg1) - erf(arg2);
      a = volume * 0.25 * tx * ty(-,);
      if( !is_void(peak) ) a *= (peak/max(a));
      return(im + a);
   } else {
      write,"Not a 2D array";
      return(-1.0);
   }
}

/* Function convol2d  */

func convol2d( a, b )
/* DOCUMENT res = convol2d(a,kernel)

   returns the convolution of 'a' and 'kernel', both 2D arrays.
*/

{

   sza = dimsof(a);
   szb = dimsof(b);
   if( (sza(1) != 2) || (szb(1) != 2) ) {
      print,"convol2d must be called with 2d arguments";
      return;
   }
   if( szb(2) % 2 != 1 || szb(3) % 2 != 1 ) {
      print,"Kernel must have odd dimensions";
      return;
   }
   ad = double(a);
   bd = double(b);
   dx = szb(2) / 2;
   dy = szb(3) / 2;

   res = ad;

   for( i = 1; i <= sza(2); i++ ) {
   for( j = 1; j <= sza(3); j++ ) {
      weight = 0.0;
      ssum = 0.0;
      for( k = -dx; k <= dx; k++ ) {
         if( k+i >= 1 && k+i <= sza(2) ) { // in large array in x
	 for( l = -dy; l <= dy; l++ ) {
	    if( l+j >= 1 && l+j <= sza(3) ) { // in large array in y
	       ssum += ad(i+k,j+l) * bd(k+dx+1,l+dy+1);
	       weight += bd(k+dx+1,l+dy+1);
	    }
	 }
	 }
      }
      res(i,j) = ssum / weight;
   }
   }
   return( res );
}

/* Function gkernel */

func gkernel( dim1, dim2, sigma)
/* DOCUMENT kernel = gkernel(dim1, dim2, sigma)

   Returns 2D kernel with 2d Gauss function
*/

{

   dim1 = (dim1/2)*2 + 1;
   dim2 = (dim2/2)*2 + 1;
   x = distances(dim1,dim2,dim1/2+1,dim2/2+1);
   kernel = exp(-0.5*(x/sigma)^2);
   return( kernel/sum(kernel) );
}

/* Function edge_suppress */

func edge_suppress( dim1, dim2, radius, lambda )
/* DOCUMENT suppress = edge_suppress( dim1, dim2, radius, lambda )

   Returns a double array dim1 x dim2 with values close to 1 when
   r < radius and approaching zero when r > radius. lambda determines
   the scale, the smaller the sharper the shift.

   2007-02-01/NJW
*/

{
   c1 = (dim1+1.0) / 2;
   c2 = (dim2+1.0) / 2;
   d = distances( dim1, dim2, c1, c2 );
   return 1.0 - sost(d, radius, lambda );
}

/* Function curmark1 */

func curmark1(void, style=, prompt=, nomark=, ps=, symsize= )
/* DOCUMENT pts = curmark1(style=, prompt=, nomark=, ps=, symsize= )
    or      curmark1, style=, prompt=, nomark=, ps=, symsize=

   Returns array with points marked with click of left mouse
   button.
   Keywords:
   'style' 0: simple click     returns (x,y)
           1: rubberband box   returns (xpres,ypres,xrelease,yrelease)
           2: rubber band line returns (xpres,ypres,xrelease,yrelease)
   'prompt': Prompt string (defaults to 'Expecting left mouse click: ...'
   'nomark': Set if point marking is to be avoided
   'ps':     Plot symbol for point mark
   'symsize': Symbol size for point mark

   When called as a subroutine the values will be printed on the screen.

   Right mouse button will return void ([]).

   SEE ALSO: curmark, mouse

*/

{
   require, "plot.i";

   local lx, ly;

   if( is_void(prompt) ) prompt = "Expecting left mouse click: ...";
   if( is_void(style) ) style = 0;
   if( is_void(ps) ) ps = 2;
   if( is_void(symsize) ) symsize = 1.;

   // Check for logarithmic scales
   mlogxy,lx,ly,get=1;

   res = mouse( -1, style, prompt );
   if( long(res(10)) != 1 ) return []; // Right mouse click
   if( lx ) res(1:3:2) = 10.^res(1:3:2);
   if( ly ) res(2:4:2) = 10.^res(2:4:2);
   if( style == 0 ) {
      if(!nomark) oplot,[res(1)],[res(2)],ps=ps,symsize=symsize;
      if( am_subroutine() ) {
         write,format="Result of curmark1: (x,y) = (%.3f,%.3f)\n", res(1), res(2);
      } else return res(1:2);
   } else {
      oplot,[res(1),res(3),res(3),res(1),res(1)], \
            [res(2),res(2),res(4),res(4),res(2)];
      if( am_subroutine() ) {
         write,format="Result of curmark1: (x1,y1) = (%.3f,%.3f)\n", res(1), res(2);
         write,format="                    (x2,y2) = (%.3f,%.3f)\n", res(3), res(4);
      } else return res(1:4);
   }
}

/* Function curmark */

func curmark( void, nomark=, ps=, symsize= )
/* DOCUMENT pts = curmark( nomark=, ps=, symsize= )
     or     curmark, nomark=, ps=, symsize=

   Returns array with points marked with click of left mouse
   button.
   Keywords:
   'nomark': Set if point marking is to be avoided
   'ps':     Plot symbol for point mark
   'symsize': Symbol size for point mark

   When called as a subroutine the values will be printed on the screen.

   SEE ALSO: curmark1, mouse

*/

{
   require, "plot.i";

   local lx, ly;

   if( is_void(ps) ) ps = 2;
   if( is_void(symsize) ) symsize = 1.;

   // Check for logarithmic scales
   mlogxy,lx,ly,get=1;

   pts = [];
   i = 0;
   while( 1 ) {
      res = mouse( -1, 0, "Expecting left mouse click (right to stop ) ..." );
      if( long(res(10)) != 1 ) return pts;
      if( lx ) res(1:3:2) = 10.^res(1:3:2);
      if( ly ) res(2:4:2) = 10.^res(2:4:2);
      if(!nomark) oplot,[res(1)],[res(2)],ps=ps,symsize=symsize;
      if( am_subroutine() ) {
         write,format="Position %3i: (x,y) = (%.3f,%.3f)\n", ++i, res(1), res(2);
      }
      grow, pts, res(1:2);
   }
}

/* Function peaksearch0 */

func peaksearch0( im, reg=, src_rad=, bkg_rad1=, bkg_rad2=, show= )
/* DOCUMENT  res = peaksearch0( im, reg=, src_rad=, bkg_rad1=, bkg_rad2=, show= )

    Returns highest pixel value, position, and total of the strongest peak
    in the image 'im' :
     [maximal value, xpos, ypos, summed peak - bkg, significance,
                                         src_rad,bkg_rad1,bkg_rad2].

    If the keyword 'reg' (for region) has been given then the search
    for a maximum is confined to im(reg).

    The keywords define the source radius (src_rad) and the
    background annulus between bkg_rad1 and bkg_rad2.
    Default values are  src_rad  =  3.0 pixels
                        bkg_rad1 =  4.0 pixels
                        bkg_rad2 = 10.0 pixels
    [The values are saved to externals _Src_rad, _Bkg_rad1, _Bkg_rad2
     so that they need not be supplied at the next call]

    Keyword 'show' will display the regions and results.

    It is assumed that the image contains no NAN values.
    A cleaning can be done with 'nan2zero' found in basic.i

    This version applies only a search for the maximum
    in the image.

    SEE ALSO: peaksearch1, peaksearch2

   2006-02-23/NJW

*/
{

   extern _Src_rad, _Bkg_rad1, _Bkg_rad2;

   local mxval, xc, yc, idx;

   sz = dimsof(im);

   // The search for a maximum

   if( is_void(reg) ) {
      // The global maximum
      maxim, im, mxval, xc, yc;
   } else {
      // the local maximum
      maxarr,im(reg),mxval,idx;
      idxx = indices(im,reg(idx(1)));
      xc = idxx(1);
      yc = idxx(2);
   }

   if( is_void(src_rad) ) {
      if( is_void(_Src_rad) ) {
         src_rad = 3.0;
         _Src_rad = 3.0;
      } else src_rad = _Src_rad;
   } else {
      _Src_rad = src_rad;
   }
   if( is_void(bkg_rad1) ) {
      if( is_void(_Bkg_rad1) ) {
         bkg_rad1 = 4.0;
         _Bkg_rad1 = 4.0;
      } else bkg_rad1 = _Bkg_rad1;
   } else {
      _Bkg_rad1 = bkg_rad1;
   }
   if( is_void(bkg_rad2) ) {
      if( is_void(_Bkg_rad2) ) {
         bkg_rad2 = 10.0;
         _Bkg_rad2 = 10.0;
      } else bkg_rad2 = _Bkg_rad2;
   } else {
      _Bkg_rad2 = bkg_rad2;
   }

   if( src_rad > bkg_rad1 ) {
      write,format="bkg_rad1 redefined to src_rad: %10.3f\n", src_rad;
      bkg_rad1 = src_rad;
   }
   if( bkg_rad1 > bkg_rad2 ) {
      write,format="bkg_rad2 redefined to 1.5*bkg_rad1: %10.3f\n", 1.5*bkg_rad1;
      bkg_rad2 = 1.5*bkg_rad1;
   }

   xarr = span(1,sz(2),sz(2))(,-:1:sz(3));
   yarr = span(1,sz(3),sz(3))(-:1:sz(2),);

   d = distances( sz(2), sz(3), xc, yc );
   ann = where( d > bkg_rad1 & d < bkg_rad2 );
   w = where( d < src_rad );
   x = xarr(w);
   y = yarr(w);
   z = im(w) - avg(im(ann));
   peakcts = sum(z);
   if( peakcts <= 0.0 ) {
      xpos = double(xc);
      ypos = double(yc);
   } else {
      xpos = sum(x*z)/peakcts;
      ypos = sum(y*z)/peakcts;
      if(abs(xpos-xc) > src_rad ) xpos = double(xc);
      if(abs(ypos-yc) > src_rad ) ypos = double(yc);
   }

   // iterate for better source position

   xc = xpos;
   yc = ypos;

   d = distances( sz(2), sz(3), xc, yc );
   ann = where( d > bkg_rad1 & d < bkg_rad2 );
   w = where( d < src_rad );
   x = xarr(w);
   y = yarr(w);
   z = im(w) - avg(im(ann));
   peakcts = sum(z);
   if( peakcts <= 0.0 ) {
      xpos = double(xc);
      ypos = double(yc);
   } else {
      xpos = sum(x*z)/peakcts;
      ypos = sum(y*z)/peakcts;
      if(abs(xpos-xc) > src_rad ) xpos = double(xc);
      if(abs(ypos-yc) > src_rad ) ypos = double(yc);
   }

   if( show ) {
      disp, im;
      circa = span(0,2*pi,100);
      circx = cos(circa);
      circy = sin(circa);
      oplot,xpos+src_rad*circx,ypos+src_rad*circy;
      oplot,xpos+bkg_rad1*circx,ypos+bkg_rad1*circy;
      oplot,xpos+bkg_rad2*circx,ypos+bkg_rad2*circy;
   }

   significance = signif(im,pos=long([xpos+0.5,ypos+0.5]));

   return [double(mxval),xpos,ypos,peakcts,significance, \
                       src_rad,bkg_rad1,bkg_rad2];
}

/* Function peaksearch1 */

func peaksearch1( im, src_rad=, bkg_rad1=, bkg_rad2= )
/* DOCUMENT  res = peaksearch1( im, src_rad=, bkg_rad1=, bkg_rad2= )

    Returns position and amplitude of the strongest peak
    in the image 'im'.

    The keywords define the source radius (src_rad) and the
    background annulus between bkg_rad1 and bkg_rad2.
    Default values are  src_rad  =  7.0 pixels
                        bkg_rad1 =  7.0 pixels
                        bkg_rad2 = 10.0 pixels

    It is assumed that the image contains no NAN values
    A cleaning can be done with 'nan2zero' found in basic.i

    This version applies only a search for the maximum
    in the image followed by an 'amoeba' gauss peak fitting.

   2006-02-23/NJW

*/
{
   require, "datafit.i";

   sz = dimsof(im);
   maxim, im, mxval, xc, yc;

   if( is_void(src_rad) ) src_rad = 7.0;
   if( is_void(bkg_rad1) ) bkg_rad1 = 7.0;
   if( is_void(bkg_rad2) ) bkg_rad2 = 10.0;

   if( src_rad > bkg_rad1 ) {
      write,format="bkg_rad1 redefined to src_rad: %10.3f\n", src_rad;
      bkg_rad1 = src_rad;
   }
   if( bkg_rad1 > bkg_rad2 ) {
      write,format="bkg_rad2 redefined to 1.5*bkg_rad1: %10.3f\n", 1.5*bkg_rad1;
      bkg_rad2 = 1.5*bkg_rad1;
   }

   d = distances( sz(2), sz(3), xc, yc );
   ann = where( d > bkg_rad1 & d < bkg_rad2 );
   w = where( d < src_rad );
   xarr = span(1,sz(2),sz(2))(,-:1:sz(3));
   yarr = span(1,sz(3),sz(3))(-:1:sz(2),);
   x = xarr(w);
   y = yarr(w);
   z = im(w);

   estimate = [mxval, xc, yc, 1.1, avg(im(ann))];
   parm = gaussfit2ds_uam( x , y, z, esti=estimate, nterm=5, chat=3);
   return parm;
}

/* Function peaksearch2 */

func peaksearch2( im )
/* DOCUMENT  res = peaksearch2( im )

    Returns position and amplitude of the strongest peak
    in the image 'im'.

    The keywords define the source radius (src_rad) and the
    background annulus between bkg_rad1 and bkg_rad2.
    Default values are  src_rad  =  7.0 pixels
                        bkg_rad1 =  7.0 pixels
                        bkg_rad2 = 10.0 pixels

    It is assumed that the image contains no NAN values
    A cleaning can be done with 'nan2zero' found in basic.i

    This version applies a Mexican hat wavelet
    in the image followed by a gauss peak fitting.

   2006-02-23/NJW

*/
{

   extern Mex_hat, Mex_fk511;

   if( is_void(Mex_hat) ) mexican_hat_ps2;
   // initializing wavelet

   sz = dimsof(im);

   print,"Convolving ...";
   fim = fconvol( im, Mex_fk511 );

   maxim, fim, mxval, xc, yc;

   d = distances( sz(2), sz(3), xc, yc );
   k = d < 7.0;
   l = d > 3.0;
   ann = where( k+l == 2 );
   w = where( d < 7.0 );
   xarr = span(1,sz(2),sz(2))(,-:1:sz(3));
   yarr = span(1,sz(3),sz(3))(-:1:sz(2),);
   x = xarr(w);
   y = yarr(w);
   z = fim(w);

   estimate = [mxval, xc, yc, 1.1, avg(im(ann))];
   parm = gaussfit2ds_uam( x , y, z, esti=estimate, nterm=5, chat=1, delta=0.03);
   return parm;
}

/* Function disp9 */

func disp9( image, off= )
/* DOCUMENT disp9, image, off=

   Display image with ds9. Works only on Uhuru
   Keyword 'off' will terminate the ds9 window
   2007-10-03/NJW
*/
{
   //+extern flag_for_ds9_window;
   require, "kfits.i";
   require, "mfits.i";

   //+if( flag_for_ds9_window && off ) {
   //+   system,"dsexit";
   //+   flag_for_ds9_window = 0;
   //+   return;
   //+}

   //+if( is_void(flag_for_ds9_window) ) {
   //+   system,"dsinit";
   //+   pause,10;
   //+   flag_for_ds9_window = 1;
   //+} else if( flag_for_ds9_window == 0 ) {
   //+   system,"dsinit";
   //+   pause,10;
   //+   flag_for_ds9_window = 1;
   //+}

   writefits,"tmp.fits", image, clobber=1;
   write,"File tmp.fits ready for display";
   //+system,"dsplay tmp.fits a";

}

/* Function image2events */

struct s_Point {
   double x;
   double y;
}

func image2events( image, crpix1, crval1, cdelt1, crpix2, crval2, cdelt2, crd= )
/* DOCUMENT xy = image2events( image, crpix1, crval1, cdelt1,
                               crpix2, crval2, cdelt2, crd= )

   returns array of struct 's_Point' with elements x and y (double).
   where x = xy.x and y = xy.y are randomized events from a count image.
   Using FITS and ds9 convention that integer image indices refer to pixel
   midpoints.
   Keyword:    crd    is the array returned by 'get_crd' and will
                      override arguments

   2008-04-07/NJW
*/
{
   dms = dimsof(image);
   if( dms(1) != 2 ) {
      write,"image2events only operates on 2D images";
      return [];
   }

   im = long(image + 0.5);
   nim = sum(im);
   if( nim < 1 ) {
      write,"image2events found no counts in the image";
      return [];
   }

   if( is_void(crpix1) ) { // assume default values
      crpix1 = crval1 = crpix2 = crval2 = 0.0;
      cdelt1 = cdelt2 = 1.0;
   }
   if( numberof(crd) == 6 ) {
      crpix1 = crd(1,1); crval1 = crd(2,1); cdelt1 = crd(3,1);
      crpix2 = crd(1,2); crval2 = crd(2,2); cdelt2 = crd(3,2);
   }

   naxis1 = dms(2);
   naxis2 = dms(3);
   xi = yi = [];
   for(j=1;j<=naxis2;j++) {
      for(i=1;i<=naxis1;i++) {
         n = im(i,j);
         if( n ) {
            grow, xi, array(i,n);
            grow, yi, array(j,n);
         }
      }
   }

   r = random(nim) - 0.5;
   p = random(nim) - 0.5;

   xy = array(s_Point,nim);

   xy.x = (xi + r - crpix1)*cdelt1 + crval1;
   xy.y = (yi + p - crpix2)*cdelt2 + crval2;

   return xy;

}

/* Function rebin */

func rebin( arr, dim1, .. )
/* DOCUMENT res = rebin( arr, dim1, .. )

   Rebin a 1 or 2-D array into new dimensions that are either
   multiples of or into the original dimensions.

   No attention to keeping level or sum (counts).

   SEE ALSO: remap

   2008-01-10/NJW
*/
{
   arr = double(arr);
   dms = dimsof(arr);
   newdms = [1,dim1];
   while( more_args() ) {
      grow, newdms, next_arg();
      newdms(1)++;
   }
   if( dms(1) != newdms(1) ) error,"REBIN dimensionality mismatch";
   if( dms(1) > 2 ) error,"REBIN number of dimensions may not exceed 2";
   // check modularity
   ndims = numberof(dms)-1;
   mod = array(long, ndims);
   for( i = 1; i <= ndims; i++ ) {
      if( newdms(i+1) == dms(i+1) ) {
         mod(i) = 1;
      } else if( newdms(i+1) < dms(i+1) ) {
         r = dms(i+1) / newdms(i+1);
         if( r*newdms(i+1) != dms(i+1) ) error,"REBIN requires integer fractions";
         mod(i) = -r;
      } else {
         r = newdms(i+1) / dms(i+1);
         if( r*dms(i+1) != newdms(i+1) ) error,"REBIN requires integer fractions";
         mod(i) =  r;
      }
   }

   // Do rebinning dimension by dimension

   xarr = array(double, newdms );
   for( i = 1; i <= newdms(2); i++ ) {
      if( mod(1) == 1 ) {
         io = i;
      } else if( mod(1) < 0 ) {
         io = indgen(-mod(1)) - (i-1)*mod(1);
      } else {
         io = (i-1)/mod(1) + 1;
      }
      if( ndims == 2 ) {
         for( j = 1; j <= newdms(3); j++ ) {
            if( mod(2) == 1 ) {
               jo = j;
            } else if( mod(2) < 0 ) {
               jo = indgen(-mod(2)) - (j-1)*mod(2);
            } else {
               jo = (j-1)/mod(2) + 1;
            }
            xarr(i,j) = sum(arr(io,jo));
         }
      } else {
         xarr(i) = sum(arr(io));
      }
   }

   return xarr;

}

/* Function remap */

func remap( arr, dim1, .. )
/* DOCUMENT res = remap( arr, dim1, .. )

   Remap a 1 or 2-D array into new dimensions of unrestricted
   values. Lowering the dimensionality may lead to an inaccurate
   result since the original array is sampled rather than
   average. (There is room for improvements).

   SEE ALSO: rebin

   2008-01-10/NJW
*/
{
   arr = double(arr);
   dms = dimsof(arr);
   newdms = [1,dim1];
   while( more_args() ) {
      grow, newdms, next_arg();
      newdms(1)++;
   }
   if( dms(1) != newdms(1) ) error,"REMAP dimensionality mismatch";
   if( dms(1) > 2 ) error,"REMAP number of dimensions may not exceed 2";
   delta = 1.0 / dms(2);
   //+ xiarr = delta/2 + (indgen(dms(2))-1)*delta;
   ndelta = 1.0 / newdms(2);
   nxiarr = ndelta/2 + (indgen(newdms(2))-1)*ndelta;
   oldi = long(nxiarr/delta) + 1;
   if( dms(1) > 1 ) {
      felta = 1.0 / dms(3);
      //+ yiarr = felta/2 + (indgen(dms(3))-1)*felta;
      nfelta = 1.0 / newdms(3);
      nyiarr = nfelta/2 + (indgen(newdms(3))-1)*nfelta;
      oldj = long(nyiarr/felta) + 1;
   }

   xarr = array( double, newdms );
   for( i = 1; i <= newdms(2); i++ ) {
      if( dms(1) > 1 ) {
         for( j = 1; j <= newdms(3); j++ ) {
            xarr(i,j) = arr(oldi(i),oldj(j));
         }
      } else {
         xarr(i) = arr(oldi(i));
      }
   }

   return xarr;

}

/* Function unwiden_image */

func unwiden_image( im, w )
/* DOCUMENT new_image = unwiden_image( im, w )

   Returns an image where the edge of width 'w' has been removed.

*/
{
   return im(1+w:-w,1+w:-w);
}

/* Function widen_image */

func widen_image( im, w, val= )
/* DOCUMENT new_image = widen_image( im, w, val= )

   Returns an image that has been enlarged by 'w' (int)
   in all directions and values have been filled out
   with the edge values of the original image ('im').
   The datatype is unchanged.

   If keyword 'val' is set then the filling will be with this.

   2009-11-04/NJW

*/
{
   dms = dimsof(im);
   v = !is_void(val);
   // create new image
   wim = array(structof(im),dms(2)+2*w,dms(3)+2*w);
   // --- fill out central part
   wim(w+1:w+dms(2),w+1:w+dms(3)) = im;
   // --- fill out left side
   if( v ) wim(1:w,w+1:w+dms(3)) = val;
   else wim(1:w,w+1:w+dms(3)) = im(1,)(-:1:w,);
   // --- fill out right side
   if(v)wim(w+dms(2)+1:0,w+1:w+dms(3)) = val;
   else wim(w+dms(2)+1:0,w+1:w+dms(3)) = im(0,)(-:1:w,);
   // --- fill out lower side
   if(v)wim(w+1:w+dms(2),1:w) = val;
   else wim(w+1:w+dms(2),1:w) = im(,1)(,-:1:w);
   // --- fill out upper side
   if(v)wim(w+1:w+dms(2),w+1+dms(3):0) = val;
   else wim(w+1:w+dms(2),w+1+dms(3):0) = im(,0)(,-:1:w);
   // --- fill out lower left corner
   if(v)wim(1:w,1:w) = val;
   else wim(1:w,1:w) = im(1,1);
   // --- fill out lower right corner
   if(v)wim(dms(2)+w+1:0,1:w) = val;
   else wim(dms(2)+w+1:0,1:w) = im(0,1);
   // --- fill out upper left corner
   if(v)wim(1:w,dms(3)+w+1:0) = val;
   else wim(1:w,dms(3)+w+1:0) = im(1,0);
   // --- fill out upper right corner
   if(v)wim(dms(2)+w+1:0,dms(3)+w+1:0) = val;
   else wim(dms(2)+w+1:0,dms(3)+w+1:0) = im(0,0);

   return wim;
}

/* Function getimval */

func getimval( im )
/* DOCUMENT image_value = getimval( im )
     or     getimval, im

   Retrieve a pixel value from an image previously displayed
   by function 'disp'.

   When called as a subroutine the indices and value will be
   printed on the screen.

   SEE ALSO: disp, curmark1, getimvals

   2010-08-06/NJW
*/
{
   pos = curmark1(prompt="Mark the pixel ... ",nomark=1);
   i = long(floor(pos(1)+0.5));
   j = long(floor(pos(2)+0.5));
   value = im(i,j);

   if( am_subroutine() ) {
      stim = structof(im);
      if( stim == long || stim == int || stim == char ) {
         write,format="(%i,%i) has value %i\n", i, j, value;
      } else if( stim == double || stim == float ) {
         if( abs(value) < 1.e-3 || abs(value) > 1.e3 ) {
            write,format="(%i,%i) has value %.5e\n", i, j, value;
         } else {
            write,format="(%i,%i) has value %.5f\n", i, j, value;
         }
      } else error,"Data type not supported";
   }

   return value;

}

/* Function getimvals */

func getimvals( im )
/* DOCUMENT image_values = getimvals( im )
     or     getimvals, im

   Retrieve several pixel values from an image previously displayed
   by function 'disp'.

   When called as a subroutine the indices and values will be
   printed on the screen.

   SEE ALSO: disp, curmark, getimval

   2010-08-06/NJW
*/
{
   vals = [];
   do {
      pos = curmark1(prompt="Mark the pixel (or right click to stop) ... ",nomark=1);
      if( is_void(pos) ) break;

      i = long(floor(pos(1)+0.5));
      j = long(floor(pos(2)+0.5));
      value = im(i,j);
      grow, vals, value;

      if( am_subroutine() ) {
         stim = structof(im);
         if( stim == long || stim == int || stim == char ) {
            write,format="(%i,%i) has value %i\n", i, j, value;
         } else if( stim == double || stim == float ) {
            if( abs(value) < 1.e-3 || abs(value) > 1.e3 ) {
               write,format="(%i,%i) has value %.5e\n", i, j, value;
            } else {
               write,format="(%i,%i) has value %.5f\n", i, j, value;
            }
         } else error,"Data type not supported";
      }

   } while( 1 );

   return vals;

}

/* Function imcut */

func imcut( im, lofac, hifac, pane= )
/* DOCUMENT rescaled_im = imcut( im, lofac, hifac, pane= )

   The highest and lowest image values are changed according
   to the following algorithm:

   A histogram of image values is produced and the peak is found
   with value vpeak.

   Moving upwards the value is found where the
   histogram has fallen to 40% of the peak, called v40p.
   All image pixels where the value exceeds vpeak + hifac*(v40p - vpeak)
   is set to this value.

   Moving downwards the value is found where the
   histogram has fallen to 40% of the peak, called v40m.
   All image pixels where the value undershoots vpeak - lofac*(vpeak - v40m)
   is set to this value.

   In many case lofac = 2 and hifac = 4 gives good results.

   2010-09-08/NJW

*/
{
   local h, x;

  /*
   *   Make histogram to decide where a sensible cut should be made
   */

   imw = double(im);
   imx = max(imw);
   imn = min(imw);
   bsize = (imx-imn)/200;
   histos, imw, h, x, binsize=bsize;
   if( !is_void(pane) ) {
      window,pane;
      plot,x,h,ps=10;
   }
   hmax = max(h);
   imax = where(hmax == h)(1);
   nh = numberof(h);
   if( imax == 1 || imax == nh ) {
      write,"IMCUT: image is not suited for this function";
      return imw;
   }
   if( hmax > 1.1 * max(h(imax-1), h(imax+1)) ) {
      hmax = h(imax) = 1.1 * max(h(imax-1), h(imax+1));
   }

   // find lower half

   for( i = imax; i > 0; i-- ) {
      if( h(i) < 0.4*hmax ) break;
   }

   // find upper half

   for( j = imax; j <= nh; j++ ) {
      if( h(j) < 0.4*hmax ) break;
   }

   if( j - i == 2 ) { // There must be one value that dominates
                      // Discard and set new value
      hmax = 1.1*max(h(i),h(j));
      if( !is_void(pane) ) {
         pause,1000;
         h(imax) = hmax;
         plot,x,h,ps=10;
      }

      // find lower half

      for( i = imax; i > 0; i-- ) {
         if( h(i) < 0.4*hmax ) break;
      }

      // find upper half

      for( j = imax; j <= nh; j++ ) {
         if( h(j) < 0.4*hmax ) break;
      }
   }

   dxhi = x(j) - x(imax);
   dxlo = x(imax) - x(i);

   // lower limit

   llim = x(imax) - lofac*dxlo;
   loreg = where( imw < llim );
   if( numberof(loreg) ) imw(loreg) = llim;

   // upper limit

   ulim = x(imax) + hifac*dxhi;
   hireg = where( imw > ulim );
   if( numberof(hireg) ) imw(hireg) = ulim;

   if( !is_void(pane) ) {
      oplot,x(i)*[1,1],[0,hmax],li=2;
      oplot,x(j)*[1,1],[0,hmax],li=2;
      oplot,llim*[1,1],[0,hmax];
      oplot,ulim*[1,1],[0,hmax];
   }

   return imw;

}

/* Function erase_source */

func erase_source( im, res )
/* DOCUMENT new_im = erase_source(im, res)

   Returns an image where the source region defined
   by 'peaksearch0' has been replaced by random
   values from the background annulus.

   res:

  [highest value, xpos, ypos, summed peak - bkg,significance,
                              src_rad,bkg_rad1,bkg_rad2].

   2010-09-20/NJW
*/
{
   dms = dimsof( im );
   d = distances(dms(2),dms(3),res(2),res(3));
   wsrc = where(d < res(6));
   wbkg = where(d > res(7) & d < res(8) );
   nsrc = numberof(wsrc);
   nbkg = numberof(wbkg);
   w = sort(random(nbkg))(1:nsrc);
   imr = im;
   imr(wsrc) = im(wbkg(w));

   return imr;
}

/* Function gflatfield */

func gflatfield( im, sigma, &gim, sel= )
/* DOCUMENT new_im = gflatfield( im, sigma[, >gim], sel= )

   The image 'im' is folded with a gaussian kernel
   with sigma = 'sigma' and then subtracted from the
   original image. The process is repeated with
   excluded source areas.
   Keyword 'sel' must be an array of same dimension
   as 'im' with 1 where pixels are useful and zero
   elsewhere.

   2007-06-19/NJW

*/
{
   require, "fconvol.i";

   di = dimsof(im);
   if( di(1) != 2 ) {
      write,"Not a 2D image";
      return [];
   }
   dosel = 0;
   if( !is_void(sel) ) {
      d = dimsof(sel);
      if( d(1) != 2 ) {
         write,"'sel' is not a 2D image";
         return [];
      }
      if( d(2) != di(2) || d(3) != di(3) ) {
         write,"'sel' is not commensurate with 'im'";
         return [];
      }
      dosel = 1;
   }
   wsel = dosel ? where( sel ) : indgen(di(2)*di(3));
   //   Avoid changing the input image
   im1 = im;
   gim = gfconvol( im, sigma );
   //    First subtraction
   diff  = im - gim;
   //    Substitute image parts that exceed 3sigma
   //    to avoid subtracting too much around sources
   rms_1 = wrms( diff(wsel) );
   q = where( abs(diff) > 3*rms_1 );
   if( numberof(q) > 0 ) {
      im1(q) = gim(q);
      gim = gfconvol( im1, sigma );
      diff = im - gim;
   }

   return diff;
}

/* Function fflatfield */

func fflatfield( im, freq, &ffim, sel= )
/* DOCUMENT new_im = fflatfield( im, freq[, >ffim], sel= )

   The image 'im' is Fourier transformed and
   frequencies above 'freq' are removed.
   'freq' is an integer or a two-element vector of
   integers where each is smaller than half of the
   corresponding dimension.
   The flattened image is then subtracted from the
   original image.
   Keyword 'sel' must be an array of same dimension
   as 'im' with 1 where pixels are useful and zero
   elsewhere.

   2010-09-20/NJW

*/
{
   dms = dimsof(im);
   o1 = dms(2)%2;
   o2 = dms(3)%2;
   if( dms(1) != 2 ) {
      write,"Not a 2D image";
      return [];
   }
   dosel = 0;
   if( !is_void(sel) ) {
      d = dimsof(sel);
      if( d(1) != 2 ) {
         write,"'sel' is not a 2D image";
         return [];
      }
      if( d(2) != dms(2) || d(3) != dms(3) ) {
         write,"'sel' is not commensurate with 'im'";
         return [];
      }
      dosel = 1;
   }
   wsel = dosel ? where( sel ) : indgen(dms(2)*dms(3));
   //   Avoid changing the input image
   freq1 = numberof(freq) > 1 ? freq(1) : freq;
   freq2 = numberof(freq) > 1 ? freq(2) : freq;
   im1 = im;
   fim = fft(im,-1);
   /*
    *   The higher frequencies are removed symmetrically
    *   so that the backwards transformation has zero
    *   imaginary part except for rounding-off errors.
    */
   i1 = 1 + freq1;
   i2 = dms(2) + 1 - freq1;
   j1 = 1 + freq2;
   j2 = dms(3) + 1 - freq2;
   fim(i1:i2,j1:j2) = 0.0;
   ffim = fft(fim,1).re/numberof(im);

   //    Subtraction
   diff  = im - ffim;
   //    Substitute image parts that exceed 3sigma
   //    to avoid subtracting too much around sources
   //+ rms_1 = wrms( diff(wsel) );
   //+ q = where( abs(diff) > 3*rms_1 );
   //+ if( numberof(q) > 0 ) {
   //+    im1(q) = gim(q);
   //+    gim = gfconvol( im1, sigma );
   //+    diff = im - gim;
   //+ }

   return diff;
}

/* Function signif */

func signif( im, pos=, src_rad=, bkg_rad1=, bkg_rad2= )
/* DOCUMENT significance = signif( im, pos=, src_rad=, bkg_rad1=, bkg_rad2= )

   Returns the significance defined as

     (im - avg(bkg))/rms(bkg)

   Keywords:    pos          a set of (i,j)
                src_rad      defaults to  3.0 pixels
                bkg_rad1     defaults to  4.0 pixels
                bkg_rad2     defaults to 10.0 pixels
    [The values are saved to externals _Src_rad, _Bkg_rad1, _Bkg_rad2
     so that they need not be supplied in the following call]

   If 'pos' is set as a two-element vector then a single value is
   returned, else the entire significance map is returned.
*/
{
   extern _Src_rad, _Bkg_rad1, _Bkg_rad2;

   dms = dimsof(im);
   im = double(im);

   if( is_void(src_rad) ) {
      if( is_void(_Src_rad) ) {
         src_rad = 3.0;
         _Src_rad = 3.0;
      } else src_rad = _Src_rad;
   } else {
      _Src_rad = src_rad;
   }
   if( is_void(bkg_rad1) ) {
      if( is_void(_Bkg_rad1) ) {
         bkg_rad1 = 4.0;
         _Bkg_rad1 = 4.0;
      } else bkg_rad1 = _Bkg_rad1;
   } else {
      _Bkg_rad1 = bkg_rad1;
   }
   if( is_void(bkg_rad2) ) {
      if( is_void(_Bkg_rad2) ) {
         bkg_rad2 = 10.0;
         _Bkg_rad2 = 10.0;
      } else bkg_rad2 = _Bkg_rad2;
   } else {
      _Bkg_rad2 = bkg_rad2;
   }

   if( numberof(pos) == 2 ) {
      d = distances(dms(2),dms(3),pos(1),pos(2));
      bkg_area = where( d >= bkg_rad1 & d < bkg_rad2 & im != 0.0 );
      if( !numberof(bkg_area) ) {
         sigim = 0.0;
      } else {
         g_bkg_avg = wavg(im(bkg_area));
         g_bkg_rms = wrms(im(bkg_area));
         sigim = (im(pos(1),pos(2)) - g_bkg_avg)/g_bkg_rms;
      }
      return sigim;
   } else {
      sigim = im;
      for( i = 1; i <= dms(2); i++ ) {
      for( j = 1; j <= dms(3); j++ ) {
         d = distances(dms(2),dms(3),i,j);
         bkg_area = where( d >= bkg_rad1 & d < bkg_rad2 & im != 0.0 );
         if( !numberof(bkg_area) ) {
            sigim(i,j) = 0.0;
         } else {
            g_bkg_avg = wavg(im(bkg_area));
            g_bkg_rms = wrms(im(bkg_area));
            sigim(i,j) = (im(i,j) - g_bkg_avg)/g_bkg_rms;
         }
      }}
      return sigim;
   }
}

/* Function cur_extract_box */

func cur_extract_box( im, over=, color=, thick= )
/* DOCUMENT box = cur_extract_box( im, over=, color=, thick= )

   Saves the part of image 'im' (previously displayed, or at least an image
   with identical or smaller dimensions) outlined by the use of the cursor.

   Keywords
     over       Causes overplotting the image with the extracted box outline
     color      Specifies overplotting color
     thick      Specifies overplotting line thickness

   2012-05-02/NJW
*/
{
   outline = curmark1(prompt="Click and draw to other corner ... ",style=2);
   x1 = long(outline(1)+0.5);
   y1 = long(outline(2)+0.5);
   x2 = long(outline(3)+0.5);
   y2 = long(outline(4)+0.5);
   if( x1 > x2 ) { x = x1; x1 = x2; x2 = x; }
   if( y1 > y2 ) { y = y1; y1 = y2; y2 = y; }

   if( over ) oplot,[x1,x2,x2,x1,x1],[y1,y1,y2,y2,y1],color=color,thick=thick;

   return im(x1:x2,y1:y2);
}

/* Function imhcut */

func imhcut( im, locut, hicut )
/* DOCUMENT cut_im = imhcut( im, locut, hicut )

    Return an image with the values cut.

    Giving an argument as void means no action.
*/
{
   res = im;

   if( !is_void(locut) ) {
      w = where( im < locut );
      if( numberof(w) ) {
         res(w) = locut;
      } else res(1,1) = locut;       
   }

   if( !is_void(hicut) ) {
      w = where( im > hicut );
      if( numberof(w) ) {
         res(w) = hicut;
      } else res(2,1) = hicut;       
   }

   return res;
}

/* Function imcolorbar */

func imcolorbar( im, scale= )
/* DOCUMENT imcolorbar, im, scale=

    Uses 'colorbar' from 'util_fr_2106.i' to draw a colorbar
    to the right of the image with max and min from the image 'im'.

    Keyword
     scale    used to scale the level values for a nice appearance
              Automatic scaling to largest value between 1 and 10
              when not given.

    2013-01-09/NJW
*/
{
   vx = max(im);
   vn = min(im);
   avn = abs(vn);
   avx = abs(vx);
   if( is_void(scale) ) {
      scale = 10.^long(floor(log10(max(avn,avx))));
   }
   if( scale == 1 ) scale = [];
   colorbar, vn, vx, levs=7, scale=scale;
}

/* Function manip_image_get */

func manip_image_get( dol_input )
/* DOCUMENT im = manip_image_get( dol_input )

     Returns the image from a FITS file extension.
     The header is copied to externals to be used by
     the 'kwds_set' function, ready to save e.g. after
     manipulation by using 'manip_image_put'.

     The WCS struct is saved as an external into: Mig_wcs.

     SEE ALSO: manip_image_put

   2013-02-21/NJW
*/
{
   extern Mig_wcs;

   im = readfits(dol_input);
   hdr = headfits(dol_input);
   fits_copy_keys, hdr, tokwds=1;
   Mig_wcs = get_wcs( hdr );
   write,"crpix1 = ", Mig_wcs.crpix1;
   write,"crpix2 = ", Mig_wcs.crpix2;
   write,"crval1 = ", Mig_wcs.crval1;
   write,"crval2 = ", Mig_wcs.crval2;
   write,"cd1_1 = ", Mig_wcs.cd1_1;
   write,"cd2_2 = ", Mig_wcs.cd2_2;
   write,"cd1_2 = ", Mig_wcs.cd1_2;
   write,"cd2_1 = ", Mig_wcs.cd2_1;
   return im;
}

/* Function manip_image_put */

func manip_image_put( outfile, im )
/* DOCUMENT manip_image_put, outfile, im

    Writes the image 'im' to a FITS file. If 'manip_image_get'
    has been used in the same Yorick session then the header
    information from the input image is copied to the output image
    extension.

    SEE ALSO: manip_image_get

   2013-02-21/NJW
*/
{
   writefits, outfile, im, clobber=1;
   write,"Done!";
}
